// Author: Pete Kirkham
// tiny quad store benchmark.
//
// test merge sort.

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <iostream>

#include "utils/string_utils.h"
#include "core/quad_sort.h"


int move_quads (unsigned int* data, unsigned int* buffer, size_t length)
{
#ifdef USE_SSE2
    __m128i*    quad_data((__m128i*)data);
    __m128i*    quad_data_end(quad_data+length);
    __m128i*    quad_buff((__m128i*)buffer);
    int total(0);

    for (int i(0); quad_data < quad_data_end; ++quad_data, ++quad_buff, ++i)
    {
        int mixed(mix(i)%length);
        total += mixed;
        _mm_store_si128(quad_buff, _mm_load_si128(quad_data));
    }
#endif
    return total;
}



int scatter_quads (unsigned int* data, unsigned int* buffer, size_t length)
{
#ifdef USE_SSE2
    __m128i*    quad_data((__m128i*)data);
    __m128i*    quad_data_end(quad_data+length);
    __m128i*    quad_buff((__m128i*)buffer);
    int total(0);

    for (int i(0); quad_data < quad_data_end; ++quad_data, ++i)
    {
        int mixed(mix(i)%length);
        total += mixed;
        _mm_store_si128(quad_buff + mixed, _mm_load_si128(quad_data));
    }
#endif
    return total;
}


// micro benchmark
int main (int narg, char** argvp)
{
    size_t num_quads((narg > 1 ? atoi(argvp[1]) : 1) * 1024);
    size_t length(4*num_quads);

    // allocate quads
    void*           _data = malloc((length*4)+16);
    void*           _buffer = new unsigned int[length+4];
    unsigned int*   data = (unsigned int*)(((size_t)_data+15)&~0xf);
    unsigned int* buffer = (unsigned int*)(((size_t)_buffer+15)&~0xf);

    std::cout << "data @ " << to_hex(data) << " [" << to_hex(_data) << ']' << std::endl;
    std::cout << "buffer @ " << to_hex(buffer) << " [" << to_hex(_buffer) << ']' << std::endl;

    // populate quads
    for (size_t index(0); index < length; index += 4)
    {
        data[index|0] = index/4 + 1;
        data[index|1] = 0;
        data[index|2] = mix(index) % length;
        data[index|3] = 0;
    }

    std::cout << num_quads << " quads." << std::endl;

    for (size_t quad_index(0); quad_index < 8; ++quad_index)
    {
        std::cout  << quad_index << "\t</" << data[quad_index<<2] << ',' << data[(quad_index<<2)|1] << ',';
        std::cout << data[(quad_index<<2)|2] << ',' << data[(quad_index<<2)|3] << "/>" << std::endl;
    }

    enum {
        no_sort,
        move_only,
        scatter_only,
        merge_sort,
        radix16_sort,
        radix256_sort,
        radix65536_sort,
        hybrid_sort
    } sort_type(merge_sort);

    size_t block_size(1024);

    for (int i(1); i < narg; ++i)
    {
        if (strcmp(argvp[i], "-no-sort") == 0)      sort_type = no_sort;
        if (strcmp(argvp[i], "-move-only") == 0)    sort_type = move_only;
        if (strcmp(argvp[i], "-scatter-only") == 0) sort_type = scatter_only;
        if (strcmp(argvp[i], "-merge") == 0)        sort_type = merge_sort;
        if (strcmp(argvp[i], "-radix") == 0)        sort_type = radix65536_sort;
        if (strcmp(argvp[i], "-radix16") == 0)      sort_type = radix16_sort;
        if (strcmp(argvp[i], "-radix256") == 0)     sort_type = radix256_sort;
        if (strcmp(argvp[i], "-radix65536") == 0)   sort_type = radix65536_sort;
        if (strcmp(argvp[i], "-hybrid") == 0)
        {
            sort_type = hybrid_sort;
            if (i+1 < narg)
                block_size = atoi(argvp[i+1]);
            break;
        }
    }

    // sort quads
    switch (sort_type)
    {
        case no_sort:
            std::cout << "not sorting:" << std::endl;
            break;
        case move_only:
            std::cout << "moving quads 10 times :" << std::endl;
            for (int i = 0; i < 10; ++i)
                move_quads(data, buffer, num_quads);
            break;
        case scatter_only:
            std::cout << "scattering quads 10 times :" << std::endl;
            for (int i = 0; i < 10; ++i)
                scatter_quads(data, buffer, num_quads);
            break;
        case merge_sort:
            std::cout << "sorting using merge sort:" << std::endl;
            data = merge_sort_quads(data, buffer, num_quads, 2);
            break;
        case radix16_sort:
            std::cout << "sorting using radix 16 sort:" << std::endl;
            data = radix16_sort_quads(data, buffer, num_quads, 2);
            break;
        case radix256_sort:
            std::cout << "sorting using radix 256 sort:" << std::endl;
            data = radix256_sort_quads(data, buffer, num_quads, 2);
            break;
        case radix65536_sort:
            std::cout << "sorting using radix 65536 sort:" << std::endl;
            data = radix65536_sort_quads(data, buffer, num_quads, 2);
            break;
        case hybrid_sort:
            std::cout << "sorting using hybrid radix 256 + merge sort:" << std::endl;
            data = hybrid_sort_quads(data, buffer, num_quads, 2, block_size);
            break;
    }

    std::cout << "sorted quads." << std::endl;
    std::cout << "checking:" << std::endl;

    for (size_t index(4); index < length; index += 4)
    {
        if (((data[index|2] < data[(index-4)|2]) || // is ordered
             ((data[index|2] == data[(index-4)|2]) && (data[index] < data[(index-4)]) )) && // sould be stable
            (sort_type != no_sort) && (sort_type != move_only) && (sort_type != scatter_only)) // and it actually tried to sort it
        {
            size_t quad_index(index / 4);

            std::cerr << "not sorted at " << quad_index << ":" << std::endl;

            size_t min(quad_index > 4 ? quad_index-4 : 0);
            size_t max(quad_index < num_quads-4 ? quad_index+4 : num_quads);

            for (quad_index = min; quad_index < max; ++quad_index)
            {
                std::cout  << quad_index << "\t</" << data[quad_index<<2] << ',' << data[(quad_index<<2)|1] << ',';
                std::cout << data[(quad_index<<2)|2] << ',' << data[(quad_index<<2)|3] << "/>" << std::endl;
            }

            free(_data);
            free(_buffer);
            exit(-1);
        }
    }

    std::cout << "checked." << std::endl;
    std::cout << std::endl;

    for (size_t quad_index(0); quad_index < 8; ++quad_index)
    {
        std::cout  << quad_index << "\t</" << data[quad_index<<2] << ',' << data[(quad_index<<2)|1] << ',';
        std::cout << data[(quad_index<<2)|2] << ',' << data[(quad_index<<2)|3] << "/>" << std::endl;
    }

    for (int i(1); i < narg; ++i)
    {
        if (strcmp(argvp[i], "-save") == 0)
        {
            // dump core
            int file(open("sorted.quads", O_WRONLY|O_CREAT, S_IRUSR|S_IWUSR | S_IRGRP|S_IWGRP));

            if (!file)
            {
                std::cout << "failed to open " << argvp[1] << std::endl;
                free(_data);
                free(_buffer);
                return 2;
            }

            write(file, data, length*4);

            close(file);
        }
    }

    free(_data);
    free(_buffer);

    return 0;
}
